package org.nakedobjects.nos.store.hibernate;

import java.util.List;

import org.apache.log4j.Logger;
import org.hibernate.Criteria;
import org.hibernate.Hibernate;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.collection.PersistentCollection;
import org.hibernate.criterion.Example;
import org.hibernate.criterion.MatchMode;
import org.hibernate.criterion.Restrictions;
import org.hibernate.engine.SessionImplementor;
import org.hibernate.metadata.ClassMetadata;
import org.hibernate.proxy.HibernateProxy;
import org.hibernate.proxy.LazyInitializer;
import org.nakedobjects.noa.adapter.NakedCollection;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.Oid;
import org.nakedobjects.noa.adapter.Persistable;
import org.nakedobjects.noa.adapter.ResolveState;
import org.nakedobjects.noa.persist.InstancesCriteria;
import org.nakedobjects.noa.persist.ObjectNotFoundException;
import org.nakedobjects.noa.persist.ObjectPersistenceException;
import org.nakedobjects.noa.persist.UnsupportedFindException;
import org.nakedobjects.noa.reflect.NakedObjectField;
import org.nakedobjects.noa.spec.Features;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.context.Perspective;
import org.nakedobjects.nof.core.context.PerspectiveCriteria;
import org.nakedobjects.nof.core.persist.AllInstances;
import org.nakedobjects.nof.core.persist.PatternCriteria;
import org.nakedobjects.nof.core.persist.TitleCriteria;
import org.nakedobjects.nof.core.util.Assert;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.persist.objectstore.NakedObjectStore;
import org.nakedobjects.nof.persist.transaction.CreateObjectCommand;
import org.nakedobjects.nof.persist.transaction.DestroyObjectCommand;
import org.nakedobjects.nof.persist.transaction.ExecutionContext;
import org.nakedobjects.nof.persist.transaction.PersistenceCommand;
import org.nakedobjects.nof.persist.transaction.SaveObjectCommand;
import org.nakedobjects.nos.store.hibernate.service.HibernateInstancesCriteria;


public class HibernateObjectStore implements NakedObjectStore {
    private static final Logger LOG = Logger.getLogger(HibernateObjectStore.class);

    private boolean isInitialized = false;

    public HibernateObjectStore() {}

    public void abortTransaction() {
        LOG.debug("Abort TX");
        HibernateUtil.rollbackTransaction();
    }

    protected boolean startHibernateTransaction() {
        Assert.assertFalse(HibernateUtil.inTransaction());
        HibernateUtil.startTransaction();
        return true;
    }

    protected void useExistingHibernateTransaction() {
        Assert.assertTrue(HibernateUtil.inTransaction());
    }

    protected void commitHibernateTransaction(final boolean started) {
        if (started) {
            Assert.assertTrue(HibernateUtil.inTransaction());
            HibernateUtil.commitTransaction();
        }
    }

    public CreateObjectCommand createCreateObjectCommand(final NakedObject object) {
        LOG.debug("  create object " + object);
        return new CreateObjectCommand() {
            public void execute(final ExecutionContext context) {
                useExistingHibernateTransaction();
                LOG.debug("  create object - actual save " + object);
                HibernateUtil.getCurrentSession().save(object.getObject());
            }

            public NakedObject onObject() {
                return object;
            }

            public String toString() {
                return "CreateObjectCommand [object=" + object + "]";
            }
        };
    }

    public DestroyObjectCommand createDestroyObjectCommand(final NakedObject object) {
        LOG.debug("  destroy object " + object);
        return new DestroyObjectCommand() {
            public void execute(final ExecutionContext context) {
                useExistingHibernateTransaction();
                LOG.debug("  destroy object - actual delete " + object);
                HibernateUtil.getCurrentSession().delete(object.getObject());
            }

            public NakedObject onObject() {
                return object;
            }

            public String toString() {
                return "DestroyObjectCommand [object=" + object + "]";
            }
        };
    }

    protected Query createQuery(final String select, final NakedObjectSpecification specification, final boolean includeSubclasses) {
        useExistingHibernateTransaction();
        final StringBuffer querySb = new StringBuffer(128);
        if (select != null) {
            querySb.append(select).append(" ");
        }
        querySb.append("from ").append(specification.getFullName()).append(" as o");
        if (!includeSubclasses) {
            final ClassMetadata meta = HibernateUtil.getSessionFactory().getClassMetadata(specification.getFullName());
            if (meta.hasSubclasses()) {
                querySb.append(" where o.class = ").append(specification.getFullName());
            }
        }
        final String queryString = querySb.toString();
        LOG.debug("Query - " + queryString);
        return HibernateUtil.getCurrentSession().createQuery(queryString);
    }

    public SaveObjectCommand createSaveObjectCommand(final NakedObject object) {
        LOG.debug("  save object " + object);
        return new SaveObjectCommand() {
            public void execute(final ExecutionContext context) {
                useExistingHibernateTransaction();
                LOG.debug("  save object - actual update " + object);
                Object toUpdate;
                if (Features.isAggregated(object.getSpecification())) {
                    final NakedObject parent = object.getSpecification().getAggregate(object);
                    LOG.debug("change to internal collection being persisted through parent");
                    toUpdate = parent.getObject();
                } else {
                    toUpdate = object.getObject();
                }
                HibernateUtil.getCurrentSession().update(toUpdate);
            }

            public NakedObject onObject() {
                return object;
            }

            public String toString() {
                return "SaveObjectCommand [object=" + object + "]";
            }
        };
    }

    public void debugData(final DebugString debug) {}

    public String debugTitle() {
        return null;
    }

    public void endTransaction() {
        LOG.debug("Committing TX");
        // do nothing - control hibernate transactions locally
    }

    protected void executeCommands(final PersistenceCommand[] commands) {
        try {
            for (int i = 0; i < commands.length; i++) {
                final PersistenceCommand command = commands[i];
                command.execute(null);
            }
        } catch (RuntimeException e) {
            LOG.warn("Failure during execution", e);
            HibernateUtil.rollbackTransaction();
            throw e;
        }
    }

    public void execute(final PersistenceCommand[] commands) {
        LOG.debug("execute " + commands.length + " commands");
        if (commands.length > 0) {
            try {
                boolean started = startHibernateTransaction();
                executeCommands(commands);
                commitHibernateTransaction(started);
            } catch (final RuntimeException e) {
                HibernateUtil.rollbackTransaction();
                throw e;
            }
        }
    }

    public boolean flush(final PersistenceCommand[] commands) {
        LOG.debug("flush " + commands.length + " commands");
        if (commands.length > 0) {
            executeCommands(commands);
            return true;
        }
        return false;
    }

    public NakedObject[] getInstancesByTitleCriteria(final TitleCriteria criteria) {
        try {
            boolean started = startHibernateTransaction();
            final String classFullName = criteria.getSpecification().getFullName();
            final Criteria hibernateCriteria = HibernateUtil.getCurrentSession().createCriteria(classFullName).add(
                    Restrictions.ilike("title", criteria.getRequiredTitle(), MatchMode.ANYWHERE));
            if (!criteria.includeSubclasses()) {
                final ClassMetadata meta = HibernateUtil.getSessionFactory().getClassMetadata(classFullName);
                if (meta.hasSubclasses()) {
                    hibernateCriteria.add(Restrictions.eq("class", classFullName));
                }
            }
            final List results = hibernateCriteria.list();
            final NakedObject[] loadedObjects = loadObjects(HibernateUtil.getCurrentSession(), criteria.getSpecification(),
                    results);
            commitHibernateTransaction(started);
            return loadedObjects;
        } catch (final RuntimeException e) {
            HibernateUtil.rollbackTransaction();
            throw e;
        }
    }

    public NakedObject[] getInstancesByHibernateCriteria(final HibernateInstancesCriteria criteria) {
        try {
            boolean started = startHibernateTransaction();
            criteria.setSession(HibernateUtil.getCurrentSession());
            final List results = criteria.getResults();
            final NakedObject[] loadedObjects = loadObjects(HibernateUtil.getCurrentSession(), criteria.getSpecification(),
                    results);
            commitHibernateTransaction(started);
            return loadedObjects;
        } catch (final RuntimeException e) {
            HibernateUtil.rollbackTransaction();
            throw e;
        }
    }

    public NakedObject[] getInstancesByPatternCriteria(final PatternCriteria criteria) {
        try {
            boolean started = startHibernateTransaction();
            Object pojoPattern = criteria.getPattern().getObject();
            // ignore title when pattern matching as it cannot be null (filled in by adapter)
            final Criteria hibernateCriteria = HibernateUtil.getCurrentSession().createCriteria(pojoPattern.getClass()).add(
                    Example.create(pojoPattern).ignoreCase().enableLike(MatchMode.ANYWHERE).excludeProperty("title"));
            if (!criteria.includeSubclasses()) {
                final ClassMetadata meta = HibernateUtil.getSessionFactory().getClassMetadata(pojoPattern.getClass().getName());
                if (meta.hasSubclasses()) {
                    hibernateCriteria.add(Restrictions.eq("class", pojoPattern.getClass()));
                }
            }
            final List results = hibernateCriteria.list();
            final NakedObject[] loadedObjects = loadObjects(HibernateUtil.getCurrentSession(), criteria.getSpecification(),
                    results);
            commitHibernateTransaction(started);
            return loadedObjects;
        } catch (final RuntimeException e) {
            HibernateUtil.rollbackTransaction();
            throw e;
        }
    }

    public NakedObject[] getPerspectiveInstances(final PerspectiveCriteria criteria) {
        try {
            boolean started = startHibernateTransaction();
            final Perspective pojoPattern = new Perspective();
            pojoPattern.setUserName(criteria.getUserName());
            // ignore title when pattern matching as it cannot be null (filled in by adapter)
            final Criteria hibernateCriteria = HibernateUtil.getCurrentSession().createCriteria(pojoPattern.getClass()).add(
                    Example.create(pojoPattern).excludeProperty("title"));
            if (!criteria.includeSubclasses()) {
                final ClassMetadata meta = HibernateUtil.getSessionFactory().getClassMetadata(pojoPattern.getClass().getName());
                if (meta.hasSubclasses()) {
                    hibernateCriteria.add(Restrictions.eq("class", pojoPattern.getClass()));
                }
            }
            final List results = hibernateCriteria.list();
            final NakedObject[] loadedObjects = loadObjects(HibernateUtil.getCurrentSession(), criteria.getSpecification(),
                    results);
            commitHibernateTransaction(started);
            return loadedObjects;
        } catch (final RuntimeException e) {
            HibernateUtil.rollbackTransaction();
            throw e;
        }
    }

    public NakedObject[] getInstances(final InstancesCriteria criteria) {
        if (criteria instanceof TitleCriteria) {
            return getInstancesByTitleCriteria((TitleCriteria) criteria);
        } else if (criteria instanceof PatternCriteria) {
            return getInstancesByPatternCriteria((PatternCriteria) criteria);
        } else if (criteria instanceof HibernateInstancesCriteria) {
            return getInstancesByHibernateCriteria((HibernateInstancesCriteria) criteria);
        } else if (criteria instanceof AllInstances) {
            return getInstances(criteria.getSpecification(), criteria.includeSubclasses());
        } else if (criteria instanceof PerspectiveCriteria) {
            return getPerspectiveInstances((PerspectiveCriteria) criteria);
        } else {
            throw new UnsupportedFindException("Can't use criteria for search: " + criteria);
        }
    }

    public NakedObject[] getInstances(final NakedObjectSpecification specification, final boolean includeSubclasses) {

        LOG.debug("getInstances - class=" + specification.getFullName() + ", subclasses=" + includeSubclasses);
        NakedObject[] objects;
        try {
            boolean started = startHibernateTransaction();
            final Query query = createQuery(null, specification, includeSubclasses);
            objects = loadObjects(HibernateUtil.getCurrentSession(), specification, query.list());
            commitHibernateTransaction(started);
        } catch (final RuntimeException e) {
            HibernateUtil.rollbackTransaction();
            throw e;
        }

        return objects;
    }

    public NakedObject getObject(final Oid oid, final NakedObjectSpecification hint) throws ObjectNotFoundException,
            ObjectPersistenceException {

        LOG.debug("getObject id=" + oid);
        Object result = null;
        final HibernateOid hoid = (HibernateOid) oid;
        try {
            boolean started = startHibernateTransaction();
            result = HibernateUtil.getCurrentSession().get(hoid.getClassName(), hoid.getPrimaryKey());
            commitHibernateTransaction(started);
        } catch (final RuntimeException e) {
            HibernateUtil.rollbackTransaction();
            throw e;
        }

        if (result == null) {
            throw new ObjectNotFoundException(oid);
        }
        return NakedObjectsContext.getObjectLoader().recreateAdapter(hoid, result);
    }

    public void loadObject(final NakedObject object) throws ObjectNotFoundException, ObjectPersistenceException {

        LOG.debug("loadObject id=" + object.getOid());
        final HibernateOid hoid = (HibernateOid) object.getOid();
        try {
            boolean started = startHibernateTransaction();
            HibernateUtil.getCurrentSession().load(hoid.getClassName(), hoid.getPrimaryKey());
            commitHibernateTransaction(started);
        } catch (final RuntimeException e) {
            HibernateUtil.rollbackTransaction();
            throw e;
        }

        if (object.getObject() == null) {
            throw new ObjectNotFoundException(object.getOid());
        }
    }

    public Oid getOidForService(final String name) {
        return new HibernateOid(name, name, name);
    }

    public boolean hasInstances(final NakedObjectSpecification specification, final boolean includeSubclasses) {
        if (specification.persistable() == Persistable.TRANSIENT) {
            LOG.warn("trying to run hasInstances for non-persistent class " + specification);
            return false;
        }
        LOG.debug("hasInstances - class=" + specification.getFullName() + ", subclasses=" + includeSubclasses);

        boolean result;
        try {
            boolean started = startHibernateTransaction();
            final Query query = createQuery("select o.id", specification, includeSubclasses);
            query.setMaxResults(1);
            result = !query.list().isEmpty();
            commitHibernateTransaction(started);
        } catch (final RuntimeException e) {
            HibernateUtil.rollbackTransaction();
            throw e;
        }

        return result;
    }

    public void init() {
        // will only return true on first execution
        isInitialized = HibernateUtil.init();
    }

    public void initializeField(final Object parent, final Object field) {
        if (!Hibernate.isInitialized(field)) {
            // TODO push this functionality up into
            // OneTo(One|Many)Transaction.get...() which should
            // wrap a transaction around all method calls which could hit the
            // db.

            // This method may not be running within the scope of a transaction,
            // so
            // make sure one is active

            try {
                boolean started = startHibernateTransaction();
                if (field instanceof PersistentCollection) {
                    HibernateUtil.getCurrentSession().lock(parent, org.hibernate.LockMode.NONE);
                    Hibernate.initialize(field);
                } else {
                    final LazyInitializer lazy = ((HibernateProxy) field).getHibernateLazyInitializer();
                    lazy.setSession((SessionImplementor) HibernateUtil.getCurrentSession());
                    lazy.initialize();
                }
                commitHibernateTransaction(started);
                LOG.debug("initializeField of type " + field.getClass() + " for parent " + parent);
            } catch (final RuntimeException e) {
                HibernateUtil.rollbackTransaction();
                throw e;
            }

        }
    }

    public boolean isInitialized() {
        return isInitialized;
    }

    protected NakedObject[] loadObjects(final Session session, final NakedObjectSpecification specification, final List list) {
        final HibernateObjectLoaderImpl loader = (HibernateObjectLoaderImpl) NakedObjectsContext.getObjectLoader();
        final NakedObject[] objects = new NakedObject[list.size()];
        int i = 0;
        for (final Object object : list) {
            // cannot just load adapter for object - if Naked Objects has
            // already loaded the object
            // then object won't match it (e.g. if getInstances has been called
            // and an instance has
            // been loaded) - so need to use Hibernate session to get an Oid to
            // do a lookup in that case
            objects[i++] = loader.getAdapterFor(object);
        }
        return objects;
    }

    public String name() {
        return "Hibernate Object Store";
    }

    public void registerService(final String service, final Oid oid) {}

    public void reset() {
        LOG.debug("resetting object store");
        // HibernateUtil.rebuildSessionFactory();
    }

    public void resolveField(final NakedObject object, final NakedObjectField field) {
        LOG.debug("resolveField id=" + object.getOid() + ", field=" + field.getId());
        if (field.isObject()) {
            final NakedObject adapter = (NakedObject) field.get(object);
            initializeField(object.getObject(), adapter.getObject());
        } else if (field.isCollection()) {
            final NakedCollection collectionAdapter = (NakedCollection) field.get(object);
            // can sometimes be called if collection is already resolved! e.g.
            // BasicPerpsective.services
            if (!collectionAdapter.getResolveState().isResolved()) {
                NakedObjectsContext.getObjectLoader().start(collectionAdapter, ResolveState.RESOLVING);
                final Object collection = collectionAdapter.getObject();
                initializeField(object.getObject(), collection);
                NakedObjectsContext.getObjectLoader().end(collectionAdapter);
            }
        }
    }

    public void resolveImmediately(final NakedObject object) {
        LOG.debug("resolveImmediately id=" + object.getOid());
        loadObject(object);
    }

    public void shutdown() {
        HibernateUtil.shutdown();
    }

    public void startTransaction() {
        LOG.debug("Start TX");
        // do nothing - control hibernate transactions locally
    }

}
// Copyright (c) Naked Objects Group Ltd.
