package org.nakedobjects.nos.store.hibernate;

import org.apache.log4j.Logger;
import org.hibernate.HibernateException;
import org.hibernate.event.InitializeCollectionEvent;
import org.hibernate.event.InitializeCollectionEventListener;
import org.hibernate.event.PostInsertEvent;
import org.hibernate.event.PostInsertEventListener;
import org.hibernate.event.PostLoadEvent;
import org.hibernate.event.PostLoadEventListener;
import org.hibernate.event.PostUpdateEvent;
import org.hibernate.event.PostUpdateEventListener;
import org.hibernate.event.PreInsertEvent;
import org.hibernate.event.PreInsertEventListener;
import org.hibernate.event.PreLoadEvent;
import org.hibernate.event.PreLoadEventListener;
import org.hibernate.event.PreUpdateEvent;
import org.hibernate.event.PreUpdateEventListener;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedObjectLoader;
import org.nakedobjects.noa.adapter.ResolveState;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
/**
 * Respond to events within Hibernate which need to be reflected within the Naked Objects
 * System.
 */
// TODO look at clearDirty - which events do we need to monitor (flush?)
public class NakedEventListener implements InitializeCollectionEventListener, PreInsertEventListener,
		PostInsertEventListener, PreLoadEventListener, PostLoadEventListener, 
		PreUpdateEventListener, PostUpdateEventListener {
	private static final long serialVersionUID = 1L;
	private final static Logger LOG = Logger.getLogger(NakedEventListener.class);
	
	public void onInitializeCollection(final InitializeCollectionEvent event) throws HibernateException {
		if (LOG.isDebugEnabled()) {
			LOG.info("InitializeCollectionEvent");
		}
		NakedObjectLoader loader = NakedObjectsContext.getObjectLoader();
		NakedObject adapter = loader.getAdapterFor(event.getCollection());
		if (adapter != null && adapter.getResolveState().isResolvable(ResolveState.RESOLVING)) {			
			if (LOG.isDebugEnabled()) {
				LOG.debug("Setting collection to resolved "+adapter);
			}
            adapter.changeState(ResolveState.RESOLVING);
            adapter.changeState(ResolveState.RESOLVED);
		}
	}

	public boolean onPreInsert(final PreInsertEvent event) {
		if (LOG.isDebugEnabled()) {
			LOG.debug("PreInsertEvent "+event.getEntity().getClass()+" "+event.getId());
		}
		NakedObject adapter = NakedObjectsContext.getObjectLoader().getAdapterFor(event.getEntity());
		if (adapter.getResolveState() == ResolveState.TRANSIENT) {
			// need to make sure object won't respond to changes in it's state
			// This may have already been done by the PersistAlgorithm, and though this is
			// probably the case it can't be guaranteed
			adapter.changeState(ResolveState.RESOLVED);
			adapter.changeState(ResolveState.UPDATING);
		}
		return false; // object not changed
	}

	public void onPostInsert(final PostInsertEvent event) {
		if (LOG.isDebugEnabled()) {
			LOG.debug("PostInsertEvent "+event.getEntity().getClass()+" "+event.getId());
		}
		// make sure the oid is loaded with the hibernate id
		// make sure the adapter is set to resolved
		NakedObjectLoader loader = NakedObjectsContext.getObjectLoader();
		NakedObject adapter = loader.getAdapterFor(event.getEntity());
		HibernateOid oid = (HibernateOid) adapter.getOid();
		if (oid.isTransient()) {
			// in case id is a property of the object, and hence not set using OidAccessor
			oid.setHibernateId(event.getId());
            // REVIEW is this the place to make the object persistent?  Is it not already done in the PersistAlgorithm
            loader.madePersistent(adapter);
		//	((HibernateObjectLoaderImpl)loader).madePersistent(adapter);
		} else {
			// not sure we should ever get here!  Check logs to see if it happens
			LOG.warn("Not transient: oid="+oid+", resolve state="+adapter.getResolveState()+" for "+event.getEntity());
		}
//		if (adapter.getResolveState() != ResolveState.RESOLVED) {
//			LOG.info("PostInsertEvent resolve state="+adapter.getResolveState()+" for "+event.getEntity());
//			adapter.changeState(ResolveState.RESOLVED);
//		}
		adapter.getSpecification().clearDirty(adapter);
	}

	/**
	 * Called before injecting property values into a newly loaded entity instance.
	 * @param event
	 */
	public void onPreLoad(final PreLoadEvent event) {
		if (LOG.isDebugEnabled()) {
			LOG.debug("PreLoadEvent "+event.getEntity().getClass()+" "+event.getId());
		}
		// Moved to NakedInterceptor.instantiate 
//		NakedObjectLoader loader = NakedObjects.getObjectLoader();
//		HibernateOid hoid = new HibernateOid(event.getEntity().getClass(), event.getId());
//		//NakedObject object = loader.recreateAdapterForPersistent(hoid, event.getEntity());
//		NakedObject adapter = ((HibernateObjectLoaderImpl)loader).recreateAdapterForPersistentOrCreateDuplicate(hoid,event.getEntity());
//		// object may already have been loaded in NakedObjects outside of the Hibernate session
//		//  - hence Hibernate loads the object, but it's already in NOF
//		if (adapter.getResolveState().isResolvable(ResolveState.RESOLVING)) {			
//			loader.start(adapter, ResolveState.RESOLVING);
//		} else {
//			LOG.info(" cannot start(RESOLVING) - state="+adapter.getResolveState());
//			loader.start(adapter, ResolveState.UPDATING);
//		}
	}

	/**
	 * Occurs after an an entity instance is fully loaded.
	 */
	public void onPostLoad(final PostLoadEvent event) {
		if (LOG.isDebugEnabled()) {
			LOG.debug("PostLoadEvent "+event.getEntity().getClass()+" "+event.getId());
		}
		NakedObjectLoader loader = NakedObjectsContext.getObjectLoader();
		// adapter already created in onPreLoad/NakedInterceptor.instantiate
		NakedObject adapter = loader.getAdapterFor(event.getEntity());

		loader.end(adapter);
		// TODO decide whether this should go into preLoad - put in NakedInterceptor.instantiate for now
		//loader.getObjectFactory().initRecreatedObject(event.getEntity());
		adapter.getSpecification().clearDirty(adapter);
	}

	public boolean onPreUpdate(final PreUpdateEvent event) {
		if (LOG.isDebugEnabled()) {
			LOG.debug("PreUpdateEvent "+event.getEntity().getClass()+" "+event.getId());
		}
		NakedObjectLoader objectLoader = NakedObjectsContext.getObjectLoader();
		NakedObject adapter = objectLoader.getAdapterFor(event.getEntity());
		// make sure objectchanged/resolve aren't fired when Hibernate accesses the object
		// TODO: if flush processing changes remove the if (leave the changeState)
		// make sure adapter not set to serializing by preflush
		if (!adapter.getResolveState().isSerializing()) {
			//adapter.changeState(ResolveState.UPDATING);
			objectLoader.start(adapter, ResolveState.UPDATING);

		}
        return false; // object not changed
	}

	public void onPostUpdate(final PostUpdateEvent event) {
		if (LOG.isDebugEnabled()) {
			LOG.debug("PostUpdateEvent "+event.getEntity().getClass()+" "+event.getId());
		}
		NakedObjectLoader objectLoader = NakedObjectsContext.getObjectLoader();
		NakedObject adapter = objectLoader.getAdapterFor(event.getEntity());
		adapter.getSpecification().clearDirty(adapter);
		// TODO: if flush processing changes remove the if (leave the changeState)
		// make sure adapter not set to serializing by preflush
		if (adapter.getResolveState() == ResolveState.UPDATING) {
//			adapter.changeState(ResolveState.RESOLVED);
			objectLoader.end(adapter);
		}
	}
	
//	public void onPostDelete(PostDeleteEvent event)  {
//		if (LOG.isDebugEnabled()) {
//			LOG.debug("PostDeleteEvent "+event.getEntity().getClass()+" "+event.getId());
//		}
//		NakedObjectLoader loader = NakedObjects.getObjectLoader();
//		NakedObject adapter = loader.getAdapterFor(event.getEntity());
//		NakedObjects.getObjectLoader().unloaded(adapter);
//	}
}
// Copyright (c) Naked Objects Group Ltd.
