package org.nakedobjects.nos.store.hibernate;

import java.io.Serializable;
import java.util.Date;

import org.apache.log4j.Logger;
import org.hibernate.CallbackException;
import org.hibernate.EmptyInterceptor;
import org.hibernate.EntityMode;
import org.hibernate.HibernateException;
import org.hibernate.type.Type;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedObjectLoader;
import org.nakedobjects.noa.adapter.ResolveState;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nos.store.hibernate.property.PropertyHelper;


/**
 * A Hibernate Interceptor to ensure process Naked Objects correctly. It does several things including
 * <ul>
 * <li>Objects cached within NOF are used instead of loading from the database</li>
 * <li>Objects are created via NOF, so container and services are injected</li>
 * <li>modified_by and modified_on properties are updated when an object is saved or updated</li>
 * </ul>
 */
public class NakedInterceptor extends EmptyInterceptor {
    private static final long serialVersionUID = 1L;
    private static Logger LOG = Logger.getLogger(NakedInterceptor.class);

    public Object getEntity(final String entityName, final Serializable id) throws CallbackException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("getEntity entityName=" + entityName + ", id=" + id);
        }
        final HibernateOid oid = new HibernateOid(entityName, id);
        final NakedObject adapter = NakedObjectsContext.getObjectLoader().getAdapterFor(oid);
        if (adapter != null) {
            LOG.debug("getEntity found naked object with same oid - " + oid);
            if (adapter.getResolveState().isGhost() || adapter.getResolveState().isPartlyResolved()
                    || adapter.getResolveState().isResolving()) {
                // change for lazy loading objects - this prevents hibernate
                // picking up the GHOST object as an already loaded entity by only returning
                // the object if it is resolved
                return null;
            }
            return adapter.getObject();
        }
        return null;
    }

    public Object instantiate(final String entityName, final EntityMode entityMode, final Serializable id)
            throws CallbackException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("instantiate entityName=" + entityName + ", id=" + id + ", mode=" + entityMode);
        }
        NakedObjectLoader loader = NakedObjectsContext.getObjectLoader();
        HibernateOid oid = new HibernateOid(entityName, id);
        NakedObjectSpecification spec = NakedObjectsContext.getReflector().loadSpecification(entityName);
        NakedObject adapter = loader.recreateAdapterForPersistent(oid, spec);
        // TODO fudge to work around problem in getEntity where entity is a subclass
        // (HibernateOid doesn't handle subclasses!)
        if (adapter.getResolveState().isResolved()) {
            loader.start(adapter, ResolveState.UPDATING);
        } else {
            loader.start(adapter, ResolveState.RESOLVING);
        }
        Object object = adapter.getObject();
        // need to set the id in case the object has an id property (if not id is held in
        // the oid, and that's taken care of above)
        try {
            HibernateUtil.getSessionFactory().getClassMetadata(entityName).setIdentifier(object, id, entityMode);
        } catch (HibernateException e) {
            throw new CallbackException("Error getting identifier property for class " + entityName, e);
        }
        return adapter.getObject();
    }

    public boolean onFlushDirty(
            final Object entity,
            final Serializable id,
            final Object[] currentState,
            final Object[] previousState,
            final String[] propertyNames,
            final Type[] types) {
        return setModified(currentState, propertyNames);
    }

    public boolean onSave(
            final Object entity,
            final Serializable id,
            final Object[] state,
            final String[] propertyNames,
            final Type[] types) {
        return setModified(state, propertyNames);
    }

    private boolean setModified(final Object[] currentState, final String[] propertyNames) {
        int updateCount = 0;
        for (int i = 0; i < propertyNames.length; i++) {
            if (propertyNames[i].equals(PropertyHelper.MODIFIED_BY)) {
                currentState[i] = NakedObjectsContext.getSession().getUserName();
                if (updateCount == 1) {
                    return true;
                }
                updateCount++;
            } else if (propertyNames[i].equals(PropertyHelper.MODIFIED_ON)) {
                currentState[i] = new Date();
                if (updateCount == 1) {
                    return true;
                }
                updateCount++;
            }
        }
        return updateCount > 0;
    }

    public Boolean isTransient(Object entity) {
        final NakedObject nakedObject = NakedObjectsContext.getObjectLoader().getAdapterFor(entity);
        if (nakedObject != null) {
            final HibernateOid hoid = (HibernateOid) nakedObject.getOid();
            return Boolean.valueOf(hoid.isTransient());
        }
        return super.isTransient(entity);
    }

}
// Copyright (c) Naked Objects Group Ltd.
