package org.nakedobjects.nos.store.hibernate;

import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.Oid;
import org.nakedobjects.noa.persist.InstancesCriteria;
import org.nakedobjects.noa.persist.ObjectNotFoundException;
import org.nakedobjects.noa.persist.ObjectPersistenceException;
import org.nakedobjects.noa.reflect.NakedObjectField;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.persist.objectstore.NakedObjectStore;
import org.nakedobjects.nof.persist.transaction.CreateObjectCommand;
import org.nakedobjects.nof.persist.transaction.DestroyObjectCommand;
import org.nakedobjects.nof.persist.transaction.PersistenceCommand;
import org.nakedobjects.nof.persist.transaction.SaveObjectCommand;


/**
 * Wraps a HibernateObjectStore to ensure objects are mapped, even if they aren't specified within the
 * properties file or picked up from related objects. This class is useful when initially setting up a
 * project, as it will tell you which entity classes aren't listed in the properties and would cause a failure
 * using the standard HibernateObjectStore.
 */
public class RemappingHibernateObjectStore implements NakedObjectStore {
    private final NakedObjectStore wrapped;

    public RemappingHibernateObjectStore(final NakedObjectStore objectStore) {
        this.wrapped = objectStore;
    }

    public CreateObjectCommand createCreateObjectCommand(final NakedObject object) {
        ensureMapped(object);
        return wrapped.createCreateObjectCommand(object);
    }

    private void ensureMapped(final NakedObject object) {
        ensureMapped(object.getSpecification());
    }

    private void ensureMapped(final NakedObjectSpecification specification) {
        HibernateUtil.ensureMapped(specification);
    }

    public DestroyObjectCommand createDestroyObjectCommand(final NakedObject object) {
        ensureMapped(object);
        return wrapped.createDestroyObjectCommand(object);
    }

    public SaveObjectCommand createSaveObjectCommand(final NakedObject object) {
        ensureMapped(object);
        return wrapped.createSaveObjectCommand(object);
    }

    public NakedObject[] getInstances(final InstancesCriteria criteria) {
        ensureMapped(criteria.getSpecification());
        return wrapped.getInstances(criteria);
    }

    public NakedObject getObject(final Oid oid, final NakedObjectSpecification hint) throws ObjectNotFoundException,
            ObjectPersistenceException {
        ensureMapped(hint);
        return wrapped.getObject(oid, hint);
    }

    public boolean hasInstances(final NakedObjectSpecification specification, final boolean includeSubclasses) {
        ensureMapped(specification);
        return wrapped.hasInstances(specification, includeSubclasses);
    }

    public String name() {
        return "RemappingHibernateObjectStore (" + wrapped.name() + ")";
    }

    public void resolveField(final NakedObject object, final NakedObjectField field) {
        ensureMapped(object);
        if (field.isObject() || field.isCollection()) {
            ensureMapped(field.getSpecification());
        }
        wrapped.resolveField(object, field);
    }

    public void resolveImmediately(final NakedObject object) {
        ensureMapped(object);
        wrapped.resolveImmediately(object);
    }

    public void abortTransaction() {
        wrapped.abortTransaction();
    }

    public void endTransaction() {
        wrapped.endTransaction();
    }

    public void execute(final PersistenceCommand[] commands) {
        wrapped.execute(commands);
    }

    public boolean flush(final PersistenceCommand[] commands) {
        return wrapped.flush(commands);
    }

    public Oid getOidForService(final String name) {
        return wrapped.getOidForService(name);
    }

    public void registerService(final String name, final Oid oid) {
        wrapped.registerService(name, oid);
    }

    public void reset() {
        wrapped.reset();
    }

    public void startTransaction() {
        wrapped.startTransaction();
    }

    public void init() {
        wrapped.init();
    }

    public boolean isInitialized() {
        return wrapped.isInitialized();
    }

    public void shutdown() {
        wrapped.shutdown();
    }

    public void debugData(final DebugString debug) {
        wrapped.debugData(debug);
    }

    public String debugTitle() {
        return wrapped.debugTitle();
    }
}
// Copyright (c) Naked Objects Group Ltd.
